@@ -1,6 +1,6 @@
 /* nethacklaunch.c
  *
- * (c)2001 M. Drew Streib <dtype@dtype.org
+ * (c)2001-3 M. Drew Streib <dtype@dtype.org>
  * This program is free software; you can redistribute it and/or modify 
  * it under the terms of the GNU General Public License as published by 
  * the Free Software Foundation; either version 2 of the License, or 
@@ -30,28 +30,19 @@
  * is very small.
  */
 
-/* IMPORTANT defines */
-
-#define SHED_UID 1031	/* the uid to shed privs to */
-#define SHED_GID 1031	/* the gid to shed privs to */
-#define MAXUSERS 64000	/* solves some preallocation issues. */
-
-#define LOC_CHROOT "/opt/nethack/nethack.dtype.org"
-#define LOC_NETHACK "/bin/nethack"
-#define LOC_DGLDIR "/dgldir/rcfiles/"
-
 /* a request from the author: please leave some remnance of
  * 'based on dgamelaunch version xxx' in any derivative works, or
  * even keep the line the same altogether. I'm probably happy 
  * to make any changes you need. */
 
-#define VERLINES 6 /* number of lines in this vanity text */
+#define VERLINES 7 /* number of lines in this vanity text */
 #define VER1 "## dgamelaunch - network console game launcher\n"
-#define VER2 "## version 1.1.6\n"
+#define VER2 "## version 1.2.6\n"
 #define VER3 "## \n"
 #define VER4 "## (c)2001-3 M. Drew Streib. This program's source is released under the GPL.\n"
 #define VER5 "## Send mail to <dtype@dtype.org> for details or a copy of the source code.\n"
-#define VER6 "## Server info is at http://dtype.org/nethack/"
+#define VER6 "## *Games on this server are recorded for in-progress viewing and playback!*\n"
+#define VER7 "## Server info is at http://dtype.org/nethack/"
 
 /* ************************************************************* */
 /* ************************************************************* */
@@ -59,15 +50,26 @@
 
 /* program stuff */
 
+#include "dgamelaunch.h"
 #include <stdlib.h>
 #include <curses.h>
 #include <crypt.h>
 #include <sys/types.h>
 #include <sys/file.h> /* for flock() */
+#include <sys/ioctl.h> /* ttyrec */
 #include <errno.h>
+#include <time.h>
 #include <dirent.h>
+#include <string.h>
+#include <sys/stat.h>
 
 extern int vi_main(int argc, char **argv);
+extern int ttyrec_main(char *);
+extern int master;
+extern int slave;
+extern struct  termios tt; 
+extern struct  winsize win;
+
 
 /* global variables */
 
@@ -77,6 +79,7 @@ char my_email[81];
 char my_pw[21];
 char my_env[1025];
 char rcfilename[80];
+char ttyrec_filename[100]; 
 
 /* preallocate this mem. bad, but ohwell. is only for pointers */
 /* makes a max number of users compiled in */
@@ -89,6 +92,148 @@ char *f_env[MAXUSERS];
 int loopcontrol1 = 0;
 int loopcontrol2 = 0;
 
+/* ************************************************************* */
+/* for ttyrec */
+
+void
+ttyrec_getmaster()
+{
+	(void) tcgetattr(0, &tt);
+	(void) ioctl(0, TIOCGWINSZ, (char *)&win);
+	if ((master = open("/dev/ptmx", O_RDWR)) < 0) {
+		exit(62);
+	}
+}
+
+/* ************************************************************* */
+
+void gen_ttyrec_filename() {
+	char spbuf[20];
+	time_t rawtime;
+	struct tm *ptm;
+
+  /* append time to filename */
+	time(&rawtime);
+	ptm = gmtime(&rawtime);
+	sprintf(spbuf,"%04i",ptm->tm_year+1900);
+	strncpy(ttyrec_filename,spbuf,4);
+	strcat(ttyrec_filename,"-");
+	sprintf(spbuf,"%02i",ptm->tm_mon+1);
+	strncat(ttyrec_filename,spbuf,2);
+	strcat(ttyrec_filename,"-");
+	sprintf(spbuf,"%02i",ptm->tm_mday);
+	strncat(ttyrec_filename,spbuf,2);
+	strcat(ttyrec_filename,".");
+	sprintf(spbuf,"%02i",ptm->tm_hour);
+	strncat(ttyrec_filename,spbuf,2);
+	strcat(ttyrec_filename,":");
+	sprintf(spbuf,"%02i",ptm->tm_min);
+	strncat(ttyrec_filename,spbuf,2);
+	strcat(ttyrec_filename,":");
+	sprintf(spbuf,"%02i",ptm->tm_sec);
+	strncat(ttyrec_filename,spbuf,2);
+	strcat(ttyrec_filename,".ttyrec");
+}
+
+/* ************************************************************* */
+
+void gen_inprogress_lock() {
+	char lockfile[130];
+	int fd;
+
+	strcpy(lockfile,LOC_INPROGRESSDIR);
+	strcat(lockfile,my_name);
+	strcat(lockfile,":");
+	strcat(lockfile,ttyrec_filename);
+
+	fd = open(lockfile,O_WRONLY|O_CREAT);
+	if (flock(fd,LOCK_EX)) exit(68);
+}
+
+/* ************************************************************* */
+
+void inprogressmenu() {
+  int i;
+	DIR *pdir;
+	struct dirent *pdirent;
+	int fd;
+	struct stat pstat;
+	char buf[20];
+	char printline[130];
+	char fullname[130];
+	char ttyrecname[130];
+	char *replacestr;
+	char *games[15];
+	time_t ctime;
+	int menuchoice;
+	char argv1[8] = "ttyplay";
+	char argv2[3] = "-p";
+	char *myargv[3];
+
+
+	do {
+  	clear();
+  	mvaddstr(1,1,VER1);
+  	mvaddstr(3,1,"The following games are in progress:");
+
+		/* clean old games and list good ones */
+		i = 0;
+		pdir = opendir(LOC_INPROGRESSDIR);
+		if (!pdir) exit(140);
+
+		while ((pdirent = readdir(pdir)) && (i<=15)) {
+			strcpy(fullname,LOC_INPROGRESSDIR);
+			strncat(fullname,pdirent->d_name,100);
+
+			fd = 0;
+			fd = open(fullname,O_RDONLY);
+			if ((fd>0) && flock(fd,LOCK_EX|LOCK_NB)) {
+
+				/* stat to check idle status */
+				strcpy(ttyrecname,LOC_TTYRECDIR);
+				strncat(ttyrecname,pdirent->d_name,100);
+				replacestr = strstr(ttyrecname,":"); 
+				if (!replacestr) exit(145);
+				replacestr[0] = '/';
+				if (stat(ttyrecname,&pstat)) exit(147);
+
+				games[i] = pdirent->d_name;
+				sprintf(printline,"%c) %s (%is idle)",i+97,pdirent->d_name
+					,time(&ctime)-pstat.st_mtime);
+				mvaddstr(5+i++,1,printline);
+			}	else {
+				unlink(fullname);
+			}
+			flock(fd,LOCK_UN|LOCK_NB);
+			close(fd);
+		}
+		
+		mvaddstr(22,1,"Watch which game? (r to refresh, q to quit) => ");
+  	refresh();
+
+		menuchoice = getch();
+
+		if ((menuchoice-97)>=0 && (menuchoice-97)<i) {
+			/* valid choice has been made */
+			strcpy(ttyrecname,LOC_TTYRECDIR);
+			strncat(ttyrecname,games[menuchoice-97],100);
+			replacestr = strstr(ttyrecname,":"); 
+			if (!replacestr) exit(145);
+			replacestr[0] = '/';
+
+			clear(); refresh();
+			endwin();
+			myargv[0] = argv1;
+			myargv[1] = argv2;
+			myargv[2] = ttyrecname;
+			if (!ttyplay_main(3,myargv)) exit(146);
+			refresh();
+		}
+
+		closedir(pdir);
+	} while (menuchoice != 'q');	
+}
+
 /* ************************************************************* */
 
 void changepw() {
@@ -131,6 +276,7 @@ void drawmenu() {
   mvaddstr(4,1,VER4);
   mvaddstr(5,1,VER5);
   mvaddstr(6,1,VER6);
+  mvaddstr(7,1,VER7);
 
   if(loggedin) {
     mvaddstr(VERLINES+2,1,"Logged in as:"); 
@@ -138,16 +284,18 @@ void drawmenu() {
     mvaddstr(VERLINES+2,15,buf);
 
     mvaddstr(VERLINES+4,1,"c) Change password");
-		mvaddstr(VERLINES+5,1,"o) Edit Option File (requires vi use)");
-    mvaddstr(VERLINES+6,1,"p) Play NetHack!");
-    mvaddstr(VERLINES+7,1,"q) Quit");
-    mvaddstr(VERLINES+9,1,"=> ");
+		mvaddstr(VERLINES+5,1,"o) Edit option file (requires vi use)");
+    mvaddstr(VERLINES+6,1,"w) Watch games in progress");
+    mvaddstr(VERLINES+7,1,"p) Play nethack!");
+    mvaddstr(VERLINES+8,1,"q) Quit");
+    mvaddstr(VERLINES+10,1,"=> ");
   } else {
     mvaddstr(VERLINES+2,1,"Not logged in.");
     mvaddstr(VERLINES+4,1,"l) Login");
     mvaddstr(VERLINES+5,1,"r) Register new user");
-    mvaddstr(VERLINES+6,1,"q) Quit");
-    mvaddstr(VERLINES+8,1,"=> ");
+    mvaddstr(VERLINES+6,1,"w) Watch games in progress");
+    mvaddstr(VERLINES+7,1,"q) Quit");
+    mvaddstr(VERLINES+9,1,"=> ");
   }
 
   refresh();
@@ -467,6 +615,15 @@ int editoptions() {
 		fprintf(rcfile,"# OPTIONS=showexp,showscore,time\n");
 		fprintf(rcfile,"# OPTIONS=!autopickup,color\n");
 		fprintf(rcfile,"# OPTIONS=catname:catfoo,dogname:dogfoo,horsename:horsefoo\n");
+		fprintf(rcfile,"#\n");
+		fprintf(rcfile,"# The following lines are necessary if you want to use the menucolors option.\n");
+		fprintf(rcfile,"# OPTIONS=menucolors\n");
+		fprintf(rcfile,"# MENUCOLOR=\" blessed \"=green\n");
+		fprintf(rcfile,"# MENUCOLOR=\" holy \"=green\n");
+		fprintf(rcfile,"# MENUCOLOR=\" uncursed \"=yellow\n");
+		fprintf(rcfile,"# MENUCOLOR=\" cursed \"=red\n");
+		fprintf(rcfile,"# MENUCOLOR=\" unholy \"=red\n");
+		fprintf(rcfile,"# MENUCOLOR=\" cursed .* (being worn)\"=orange&underline\n");
 	}
 
 	fclose(rcfile);
@@ -540,6 +697,15 @@ main (int argc, char **argv)
   char inbuf[1024]; /* always check for overruns */
   int userchoice;
 
+	/* get master tty just before chroot (lives in /dev) */
+	ttyrec_getmaster();
+	grantpt( master);
+	unlockpt(master);
+	if ((slave = open((const char *)ptsname(master), O_RDWR)) < 0) {
+		exit(65);
+	}
+
+
   /* chroot */
   if (chroot (LOC_CHROOT))
     error (1, errno, "cannot change root directory");
@@ -557,12 +723,15 @@ main (int argc, char **argv)
   if (readfile(0)) exit(110);
 
   initncurses();
-  while ((userchoice!='p')||(!loggedin)) {
+  while ((userchoice!='p')|(!loggedin)) {
     drawmenu();
     userchoice = getch();
     if (userchoice=='c') {
       changepw();
     }
+    if (userchoice=='w') {
+			inprogressmenu();
+		}
 		if (userchoice=='o' && loggedin) {
 			editoptions();
 		}
@@ -586,13 +755,19 @@ main (int argc, char **argv)
 	strcat(atrcfilename,rcfilename);
 	setenv("NETHACKOPTIONS",atrcfilename,1);
 
+	/* lock */
+	gen_ttyrec_filename();
+	gen_inprogress_lock(); 
+
   /* launch program */
-  myargv[0] = argv1;
+	ttyrec_main(my_name);
+  /* myargv[0] = argv1;
   myargv[1] = argv2;
   myargv[2] = my_name;
   myargv[3] = 0;
   execvp (LOC_NETHACK, myargv);
   error (1, errno, "cannot execute %s", argv[0]);
+	*/
 
   exit (1);
   return 1;
